/*
 * This source file is part of the bstring string library.  This code was
 * written by Paul Hsieh in 2002-2003, and is covered by the BSD open source 
 * license. Refer to the accompanying documentation for details on usage and 
 * license.
 */

/*
 * bstrwrap.c
 *
 * This file is the C++ wrapper for the bstring functions.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "bstrwrap.h"

CBString::CBString (const void * blk, int len) { 
	tagbstring x;
	x.slen = x.mlen = len;
	x.data = (unsigned char *) blk;
	*this = x;
}

CBString::CBString (char c, int len) {
	slen = 0;
	mlen = len;
	data = (unsigned char *) malloc (len);
	if (BSTR_ERR == bsetstr (this, len, NULL, c)) {
		bstringThrow ("Failure in (char, int) constructor");
	}
}

CBString::~CBString () {
	if (data != NULL) {
		free (data);
		data = NULL;
	}
	mlen = 0;
	slen = -1;
}

const CBString& CBString::operator = (char c) {
	slen = sizeof (char);
	mlen = sizeof (char) * 2;
	data = (unsigned char *) malloc (sizeof (char) * 2);
	data[0] = c;
	data[1] = '\0';
	return *this;
}

const CBString& CBString::operator = (unsigned char c) {
	slen = sizeof (char);
	mlen = sizeof (char) * 2;
	data = (unsigned char *) malloc (sizeof (char) * 2);
	data[0] = c;
	data[1] = '\0';
	return *this;
}

const CBString& CBString::operator = (int c) {
	char buff[20];
	itoa (c, buff, 10);
	*this = (char *)buff;
	return *this;
}

const CBString& CBString::operator = (float c) {
	char buff[20];
	sprintf (buff, "%f", c);
	*this = (char *)buff;
	return *this;
}

const CBString& CBString::operator = (double c) {
	char buff[30];
	sprintf (buff, "%g", c);
	*this = (char *)buff;
	return *this;
}

const CBString& CBString::operator = (const char *s) {
	if (NULL == s) {
		s = "";
	}
	slen = strlen (s);
	mlen = slen + 1;
	data = (unsigned char *) malloc (mlen);
	memcpy (data, s, mlen);
	data[slen] = '\0';
	return *this;
}

const CBString& CBString::operator = (const CBString& b) {
	slen = b.slen;
	mlen = slen + 1;
	data = (unsigned char *) malloc (mlen);
	memcpy (data, b.data, mlen);
	data[slen] = '\0';
	return *this;
}

const CBString& CBString::operator = (const tagbstring& x) {
	slen = x.slen;
	mlen = slen + 1;
	data = (unsigned char *) malloc (mlen);
	memcpy (data, x.data, mlen);
	data[slen] = '\0';
	return *this;
}

const CBString& CBString::operator += (const CBString& b) {
	if (BSTR_ERR == bconcat (this, (const bstring)&b)) {	// BOZO -- why is this cast necessary?
		bstringThrow ("Failure in concatenate");
	}
	return *this;
}

const CBString& CBString::operator += (const char *s) {
	struct tagbstring x;
	x.slen = strlen (s);
	x.mlen = x.slen + 1;
	x.data = (unsigned char *)s;
	if (BSTR_ERR == bsetstr (this, this->slen, &x, '?')) {
		bstringThrow ("Failure in concatenate");
	}
	return *this;
}

const CBString& CBString::operator += (char c) {
	struct tagbstring x;
	x.slen = sizeof (char);
	x.mlen = sizeof (char) * 2;
	x.data = (unsigned char *)&c;
	if (BSTR_ERR == bsetstr (this, this->slen, &x, '?')) {
		bstringThrow ("Failure in concatenate");
	}
	return *this;
}

const CBString& CBString::operator += (unsigned char c) {
	struct tagbstring x;
	x.slen = sizeof (char);
	x.mlen = sizeof (char) * 2;
	x.data = (unsigned char *)&c;
	if (BSTR_ERR == bsetstr (this, this->slen, &x, '?')) {
		bstringThrow ("Failure in concatenate");
	}
	return *this;
}

const CBString& CBString::operator += (int c) {
	CBString rvalue (c);
	*this += rvalue;
	return *this;
}

const CBString& CBString::operator += (float c) {
	CBString rvalue (c);
	*this += rvalue;
	return *this;
}

const CBString& CBString::operator += (double c) {
	CBString rvalue (c);
	*this += rvalue;
	return *this;
}

const CBString CBString::operator + (char c) {
	CBString retval (*this);
	retval += c;
	return retval;
}

const CBString CBString::operator + (const CBString& b) const {
	CBString retval (*this);
	retval += b;
	return retval;
}

const CBString CBString::operator + (const char *s) const {
	CBString retval (*this);
	retval += s;
	return retval;
}

bool CBString::operator == (const CBString& b) const {
	int retval;
	if (BSTR_ERR == (retval = biseq ((const bstring)this, (const bstring)&b))) { // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in compare (==)");
	}
	return retval != 0;
}

bool CBString::operator != (const CBString& b) const {
	return ! ((*this) == b);
}

bool CBString::operator < (const CBString& b) const {
	int retval;
	if (BSTR_ERR == (retval = bstrcmp ((const bstring)this, (const bstring)&b))) { // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in compare (<)");
	}
	return retval < 0;
}

bool CBString::operator <= (const CBString& b) const {
	int retval;
	if (BSTR_ERR == (retval = bstrcmp ((const bstring)this, (const bstring)&b))) { // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in compare (<=)");
	}
	return retval <= 0;
}

bool CBString::operator > (const CBString& b) const {
	int retval;
	if (BSTR_ERR == (retval = bstrcmp ((const bstring)this, (const bstring)&b))) { // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in compare (>)");
	}
	return retval > 0;
}

bool CBString::operator >= (const CBString& b) const {
	int retval;
	if (BSTR_ERR == (retval = bstrcmp ((const bstring)this, (const bstring)&b))) { // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in compare (>=)");
	}
	return retval >= 0;
}

void CBString::format (const char * fmt, ...) {
	bstring b;
	va_list arglist;

	va_start (arglist, fmt);
	b = bvformat (fmt, arglist);
	va_end (arglist);
	*this = *b;
	bdestroy (b);
}

void CBString::vformat (const char * fmt, va_list arglist) {
	bstring b;

	b = bvformat (fmt, arglist);
	*this = *b;
	bdestroy (b);
}

int CBString::find (const CBString& b, int pos) const {
	return binstr ((const bstring)this, pos, (const bstring)&b);	 // BOZO -- why is this cast necessary?
}

int CBString::reversefind (const CBString& b, int pos) const {
	return binstrr ((const bstring)this, pos, (const bstring)&b);	 // BOZO -- why is this cast necessary?
}

int CBString::findchr (const CBString& b, int pos) const {
	return binchr ((const bstring)this, pos, (const bstring)&b);	 // BOZO -- why is this cast necessary?
}

int CBString::reversefindchr (const CBString& b, int pos) const {
	return binchrr ((const bstring)this, pos, (const bstring)&b);	 // BOZO -- why is this cast necessary?
}

const CBString CBString::midstr (int left, int len) const {
	bstring b = bmidstr ((const bstring)this, left, len);	 // BOZO -- why is this cast necessary?
	if (NULL == b) {
		bstringThrow ("Failure in midstr");
	}
	CBString retval (*b);
	bdestroy (b);
	return retval;
}

void CBString::alloc (int length) {
	if (BSTR_ERR == balloc ((bstring)this, length)) {
		bstringThrow ("Failure in alloc");
	}
}

void CBString::fill (int length, unsigned char fill) {
	slen = 0;
	if (BSTR_ERR == bsetstr (this, length, NULL, fill)) {
		bstringThrow ("Failure in fill");
	}
}

void CBString::setsubstr (int pos, const CBString& b, unsigned char fill) {
	if (BSTR_ERR == bsetstr (this, pos, (const bstring)&b, fill)) {	 // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in setstr");
	}
}

void CBString::insert (int pos, const CBString& b, unsigned char fill) {
	if (BSTR_ERR == binsert (this, pos, (const bstring)&b, fill)) {	 // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in insert");
	}
}

void CBString::insertchrs (int pos, int len, unsigned char fill) {
	if (BSTR_ERR == binsertch (this, pos, len, fill)) {	 // BOZO -- why is this cast necessary?
		bstringThrow ("Failure in insertchrs");
	}
}


void CBString::remove (int pos, int len) {
	if (BSTR_ERR == bdelete (this, pos, len)) {
		bstringThrow ("Failure in remove");
	}
}

void CBString::trunc (int len) {
	if (len < 0) {
		bstringThrow ("Failure in trunc");
	}
	if (len < slen) {
		slen = len;
		data[len] = '\0';
	}
}

void CBString::repeat (int count) {
	count *= slen;
	if (count == 0) {
		trunc (0);
		return;
	}
	if (count < 0 || BSTR_ERR == bpattern (this, count)) {
		bstringThrow ("Failure in repeat");
	}
}

int CBString::gets (bNgetc getcPtr, void * parm, char terminator) {
	bstring b = bgets (getcPtr, parm, terminator);
	if (b == NULL) {
		slen = 0;
		return -1;
	}
	*this = *b;
	bdestroy (b);
	return 0;
}

int CBString::read (bNread readPtr, void * parm) {
	bstring b = bread (readPtr, parm);
	if (b == NULL) {
		slen = 0;
		return -1;
	}
	*this = *b;
	bdestroy (b);
	return 0;
}

int CBString::bStreamgets (struct bStream * s, char terminator) {
	return bsreadln ((bstring) this, s, terminator);
}

int CBString::bStreamread (struct bStream * s, int n) {
	return bsread ((bstring) this, s, n);
}

int CBString::bStreamunread (struct bStream * s) {
	return bsunread (s, (bstring) this);
}

int CBString::bStreampeek (struct bStream * s) {
	return bspeek ((bstring) this, s);
}

#if defined(BSTRLIB_CAN_USE_STL)
void CBStringList::split (const CBString& b, char splitChar) {
int p, i;

	p = 0;
	do {
		for (i = p; i < b.length (); i++) {
			if (b.character (i) == splitChar) break;
		}
		if (i > p) this->push_back (CBString (&(b.data[p]), i - p));
		p = i + 1;
	} while (p < b.length ());
}

void CBStringList::split (const CBString& b, const CBString& s) {
struct { unsigned long content[256 / 32]; } chrs;
unsigned char c;
int p, i;

	if (s.length() == 0) bstringThrow ("Null splitstring failure");
	if (s.length() == 1) {
		this->split (b, s.character (0));
	} else {

		for (i=0; i < (256 / 32); i++) chrs.content[i] = 0x0;
		for (i=0; i < s.length(); i++) {
			c = s.character (i);
			chrs.content[c >> 5] |= ((long)1) << (c & 31);
		}

		p = 0;
		do {
			for (i = p; i < b.length (); i++) {
				c = b.character (i);
				if (chrs.content[c >> 5] & ((long)1) << (c & 31)) break;
			}
			if (i > p) this->push_back (CBString (&(b.data[p]), i - p));
			p = i + 1;
		} while (p < b.length ());
	}
}
#endif
