#include <dir.h>
#include <stdio.h>
#include <ctype.h>
#include <alloc.h>
#include "bstrlib.h"
#include "glob2.h"

/* Perform REN/FC compatible file globbing */

struct wildFrag {
    int len;
    char * entry;
};

/*  Given a known number and preallocates space for wild card fragments, fill
 *  in those fragments.
 */
static int matchGlobFragsFill (const char * pathMask, char * fileName, int n, struct wildFrag * wf) {
int i, j, k;

    for (i=0; pathMask[i] != '\0' && fileName[i] != '\0'; i++) {
        if (toupper (pathMask[i]) != toupper (fileName[i])) {
            switch (pathMask[i]) {
                case '?': 
                    if (n == 0) return -__LINE__;
                    wf->len = 1;
                    wf->entry = fileName + i;
                    wf++;
                    n--;
                    break;
                case '*':
                    if (n == 0) return -__LINE__;
                    while (pathMask[i] == '*') pathMask++;
                    wf->entry = fileName + i;
                    for (j=0; fileName[i+j]; j++) {
                        wf->len = j;
                        k = matchGlobFragsFill (pathMask+i, fileName + i + j, n-1, wf+1);
                        if (k == 0) return 0;
                    }
                    if (pathMask[i] == '\0' && n == 1) {
                        wf->len = j;
                        return 0;
                    }

                    /* fall through */
                default:
                    return -__LINE__;
            }
        }
    }

    if (pathMask[i] == '\0' && fileName[i] != '\0') return -__LINE__;
    if (n == 0) return 0;

    if (n == 1 && fileName[i] == '\0' && pathMask[i] == '*') {
        wf->len = 0;
        wf->entry = fileName + i;
        return 0;
    }
    return -__LINE__;
}

/*  wf and n are filled in with the fragments and number of fragments 
 *  respectively.
 */
static int matchGlobFrags (const char * pathMask, char * fileName, int * n, struct wildFrag ** wf) {
int i, c;
struct wildFrag * w;

    if (pathMask == NULL || fileName == NULL || n == NULL || wf == NULL) return -__LINE__;

    for (c=i=0; pathMask[i]; i++) {
        switch (pathMask[i]) {
            case '?': c++; break;
            case '*': while (pathMask[i+1] == '*') i++; c++; break;
        }
    }

    if (c == 0) {
        *n = 0;
        *wf = NULL;
        return 0;
    }

    w = (struct wildFrag *) malloc (c * sizeof (struct wildFrag));
    i = matchGlobFragsFill (pathMask, fileName, c, w);
    if (i) {
        free (w);
        return i;
    }
    *n = c;
    *wf = w;
    return 0;
}

/*  For a given wild card mask, and filename which matches it, an attempt is
 *  made to translate that matched file to the corresponding file which is
 *  matched by another mask by matching up the * and ? fragments.  A bstring
 *  containing filename matching the second path is returned if such a 
 *  matching is possible, otherwise NULL is returned.
 */
static bstring globExp2 (const char * pathMask1, char * fileName, const char * pathMask2) {
bstring b;
int n, i, j;
struct wildFrag * wf;
struct tagbstring t;
char * q, * p;

    if (pathMask1 == NULL || fileName == NULL || pathMask2 == NULL) {
        return NULL;
    }

    for (j=i=0; pathMask1[i]; i++) {
        if (pathMask1[i] == ':' || pathMask1[i] == '\\') j = i+1;
    }
    pathMask1 += j;

    i = matchGlobFrags (pathMask1, fileName, &n, &wf);
    if (i < 0 || wf == NULL) return NULL;

    b = cstr2bstr (pathMask2);
    if (n == 0 || b == NULL) return b;

    q = NULL;

    for (j=i=0; i < b->slen; i++) {
        start:;
        switch (b->data[i]) {
            case '?': 
                if (j >= n) {
                    goto tail;
                }
                if (wf[j].len != 1) {
                    free (wf);
                    return NULL;
                }
                b->data[i] = *wf[j].entry;
                j++;
                break;
            case '*': 
                if (j >= n) {
                    tail:;
                    p = wf[n-1].entry + wf[n-1].len;
                    free (wf);
                    if (q) {
                        bstring s;
                        int x = 0;
                        for (j=0; q[j] != 0; j++) if (q[j]=='*' || q[j]=='?') x = j+1;
                        blk2tbstr (t, q, x);
                        s = bstrcpy (&t);
                        j = matchGlobFrags (s->data, p, &n, &wf);
                        if (j >= 0 && wf != NULL) {
                            j = 0;
                            goto start;
                        }
                    }
                    return NULL;
                }
                bdelete (b, i, 1);
                blk2tbstr (t, wf[j].entry, wf[j].len);
                if (binsert (b, i, &t, '?') < 0) {
                    free (wf);
                    return NULL;
                }
                i += wf[j].len - 1;
                j++;
                break;
        }
        if (q == NULL && j >= n) {
            q = b->data + i + 1;
        }
    }
    free (wf);
    return b;
}

/****************************************************************************/

int useLFN = 0;

static struct ffblk SFNblk;

#include <dos.h>

static int findFirstFB (char * path, struct LFNFileBlk * lfnBlk, int modeMust, int modeAllow) {
  if (path == NULL || lfnBlk == NULL) return -__LINE__;
  modeAllow &= 0x3f;

  if (useLFN) {
    struct SREGS s;
    union REGS r;

    r.h.ch = modeMust;
    r.h.cl = modeAllow;
    r.x.si = 1;
    r.x.ax = 0x714E; /* long filename - find first matching file */
    r.x.dx = FP_OFF (path);
    s.ds   = FP_SEG (path);
    r.x.di = FP_OFF (lfnBlk);
    s.es   = FP_SEG (lfnBlk);
    r.x.cflag = 1;
    int86x (0x21, &r, &r, &s);

    if (r.x.cflag != 0) return -__LINE__;

    if (lfnBlk->shortFileName[0] == '\0') {
      strncpy (lfnBlk->shortFileName, lfnBlk->longFileName, 13);
    }

    lfnBlk->xtra.lfn.iterationHandle = r.x.ax;
    lfnBlk->xtra.lfn.unicodeProblems = r.x.cx;

  } else {
    int rv;
    if (path == NULL || lfnBlk == NULL) return -__LINE__;
    rv = findfirst (path, &SFNblk, modeMust & 0x3f);
    if (rv != 0) return rv;
    memset (lfnBlk, 0, sizeof (struct LFNFileBlk));
    memcpy (lfnBlk->xtra.sfn.ff_reserved, SFNblk.ff_reserved, 22);
    lfnBlk->lfn_attrib                  = SFNblk.ff_attrib;
    lfnBlk->lfn_creationTime[0]         = SFNblk.ff_ftime;
    lfnBlk->lfn_creationTime[1]         = SFNblk.ff_fdate;
    lfnBlk->lfn_lastAccessTime[0]       = SFNblk.ff_ftime;
    lfnBlk->lfn_lastAccessTime[1]       = SFNblk.ff_fdate;
    lfnBlk->lfn_lastModificationTime[0] = SFNblk.ff_ftime;
    lfnBlk->lfn_lastModificationTime[1] = SFNblk.ff_fdate;
    lfnBlk->lfn_fileSize[1]             = SFNblk.ff_fsize;
    strcpy (lfnBlk->shortFileName, SFNblk.ff_name);
  }

  return 0;
}

static int findNextFB (unsigned short iterationHandle, struct LFNFileBlk * lfnBlk) {
  if (lfnBlk == NULL) return -__LINE__;

  if (useLFN) {
    struct SREGS s;
    union REGS r;
    r.x.bx = iterationHandle;
    r.x.si = 1;
    r.x.ax = 0x714F; /* long filename - find next matching file */
    r.x.di = FP_OFF (lfnBlk);
    s.es   = FP_SEG (lfnBlk);
    r.x.cflag = 1;
    int86x (0x21, &r, &r, &s);
    if (r.x.cflag != 0) return -__LINE__;
    if (lfnBlk->shortFileName[0] == '\0') {
      strncpy (lfnBlk->shortFileName, lfnBlk->longFileName, 13);
    }
    lfnBlk->xtra.lfn.iterationHandle = iterationHandle;
    lfnBlk->xtra.lfn.unicodeProblems = r.x.cx;
  } else {
    int rv;
    memcpy (SFNblk.ff_reserved, lfnBlk->xtra.sfn.ff_reserved, 22);
    rv = findnext (&SFNblk);
    if (rv != 0) return rv;
    memset (lfnBlk, 0, sizeof (struct LFNFileBlk));
    memcpy (lfnBlk->xtra.sfn.ff_reserved, SFNblk.ff_reserved, 22);
    lfnBlk->lfn_attrib                  = SFNblk.ff_attrib;
    lfnBlk->lfn_creationTime[0]         = SFNblk.ff_ftime;
    lfnBlk->lfn_creationTime[1]         = SFNblk.ff_fdate;
    lfnBlk->lfn_lastAccessTime[0]       = SFNblk.ff_ftime;
    lfnBlk->lfn_lastAccessTime[1]       = SFNblk.ff_fdate;
    lfnBlk->lfn_lastModificationTime[0] = SFNblk.ff_ftime;
    lfnBlk->lfn_lastModificationTime[1] = SFNblk.ff_fdate;
    lfnBlk->lfn_fileSize[1]             = SFNblk.ff_fsize;
    strcpy (lfnBlk->shortFileName, SFNblk.ff_name);
  }

  return 0;
}

static int findCloseFB (unsigned short iterationHandle) {
struct SREGS s;
union REGS r;

  if (useLFN) {
    r.x.bx = iterationHandle;
    r.x.ax = 0x71A1; /* long filename - find close file */
    r.x.cflag = 1;
    int86x (0x21, &r, &r, &s);
    if (r.x.cflag != 0) return -__LINE__;
  }
  return 0;
}

int getShortFileNamePath (char * outfn, const char * infn) {
struct SREGS s;
union REGS r;

  r.x.cx = 0x8001; /* leave the subst'ed drive letters the way they are */
  /* (long filename) get short filename or path */
  if (useLFN) r.x.ax = 0x7160;
  else r.x.ax = 0x6000;
  r.x.si = FP_OFF ((char *) infn);
  s.ds   = FP_SEG ((char *) infn);
  r.x.di = FP_OFF (outfn);
  s.es   = FP_SEG (outfn);
  int86x (0x21, &r, &r, &s);
  if (r.x.cflag != 0) return -r.x.ax;
  return 0;
}

int twoPathMasksWalkExpand (char * pathMask1, const char * pathMask2, 
    int (* cb) (void * parm, const char * f0, const char * f1), void * parm) {
struct LFNFileBlk lfnBlk;
bstring x;
char * s;
int ret, c;

    if (pathMask1 == NULL) return -__LINE__;
    ret = findFirstFB (pathMask1, &lfnBlk, 0, (FA_HIDDEN | FA_SYSTEM | FA_LABEL));
    if (ret < 0) return ret;

    c = 0;
    do {
        s = lfnBlk.longFileName;
        if (s[0] == '\0') s = lfnBlk.shortFileName;
        x = globExp2 (pathMask1, s, pathMask2);
        if (x) {
            ret = cb (parm, s, x->data);
            bdestroy (x);
            if (ret < 0) {
                c = ret;
                break;
            }
            c++;
        }
    } while (findNextFB (lfnBlk.xtra.lfn.iterationHandle, &lfnBlk) == 0);
    findCloseFB (lfnBlk.xtra.lfn.iterationHandle);

    return c;
}

int onePathMasksWalkExpand (char * pathMask, 
    int (* cb) (void * parm, const char * f0), void * parm) {
struct LFNFileBlk lfnBlk;
char * s;
int ret, c = 0;

    if (pathMask == NULL) return -__LINE__;
    ret = findFirstFB (pathMask, &lfnBlk, 0, (FA_HIDDEN | FA_SYSTEM | FA_LABEL));
    if (ret < 0) return ret;

    c = 0;
    do {
        s = lfnBlk.longFileName;
        if (s[0] == '\0') s = lfnBlk.shortFileName;
        cb (parm, s);
        if (ret < 0) {
            c = ret;
            break;
        }
        c++;
    } while (findNextFB (lfnBlk.xtra.lfn.iterationHandle, &lfnBlk) == 0);
    findCloseFB (lfnBlk.xtra.lfn.iterationHandle);

    return c;
}

/****************************************************************************/

/* Detect MS-Windows */

unsigned int GetMSWin9XVer (void) {
  union REGS r;
  r.x.ax = 0x1600;
  int86 (0x2f, &r, &r);
  if ((r.h.al & 0x7f) == 0) return 0; /* 0x80 is OS/2 apparently */
  return r.h.ah + ((int)r.h.al << 8);
}

#if 0
unsigned long GetDosVer (void) {
  union REGS r;
  r.x.ax = 0x3306;
  int86 (0x21, &r, &r);
  return ((unsigned long) r.x.bx) << 16 + r.x.dx;
}

#define WINNT_DOS_VER (0x3205)

int InWindowsNT (void) {
  return GetMSWin9XVer () == 0 && (GetDosVer () >> 16) == WINNT_DOS_VER;
}
#endif

/****************************************************************************/
