#include <stdio.h>
#include <string.h>
#include <alloc.h>

char * resources = NULL;

int resourceTranslate (char * prefix, FILE * out) {
  char *r;

  if (!resources) { /* load resource file if not already done */
    return -__LINE__;
  }

  r = resources;
  while (r) {
    if (r[0] == ':') {
      char buf1;
      char * endp = strstr(r, "\n.\n"); /* find next separator */
      if (!endp) {
        endp = strstr(r, "\r\n.\r\n");
        if (!endp) return -__LINE__; /* unterminated string */
      }
      r++; /* skip over ":" */

      { /* found OUR string */
	char *startp;
        int i, col;

	startp = strchr(r,'\n');
        if (*--endp != '\\')
          endp = strchr(endp,'.'); /* include newline */
          /* else: \\ already got stripped during check */
        if (!startp) break;
        startp++; /* that is OUR string */

        for (i=0; r[i]; i++) {
            if (r[i] == '#' || r[i] == '\n' || r[i] == '\r' || r[i] == '\0')
                break;
        }

        /* if (!r[i]) return  -__LINE__; */
        buf1 = r[i];
        r[i] = '\0';
	fprintf (out, "#define %s%s \\\n\t\"", prefix, r);
        r[i] = buf1;
	col = 10;

	buf1 = endp[0];
        endp[0] = '\0'; /* MODIFY resources in place! */

        for (i=0; startp[i]; i++) {
          int c = startp[i];
          if (c == '\\') {
            if (col > 74) {
              fprintf(out, "\" \\\n\t\"");
              col = 10;
            }
            fprintf(out, "\\\\");
            col+=2;
          } else if (c > 31 && c < 127) {
            if (col > 75) {
              fprintf(out, "\" \\\n\t\"");
              col = 10;
            }
            fputc (c, out);
            col++;
          } else if (c == '\t') {
            if (col > 74) {
              fprintf(out, "\" \\\n\t\"");
              col = 10;
            }
            fprintf(out, "\\t");
            col+=2;
          } else if (c == '\r') {
            if (col > 74) {
              fprintf(out, "\" \\\n\t\"");
              col = 10;
            }
            fprintf(out, "\\r");
            col+=2;
          } else if (c == '\n') {
            if (col > 74) {
              fprintf(out, "\" \\\n\t\"");
              col = 10;
            }
            fprintf(out, "\\n");
            col+=2;
          } else {
            int d0, d1, d2;

            if (col > 71) {
              fprintf(out, "\" \\\n\t\"");
              col = 10;
            }

            d0 = c / 64;
            d1 = (c / 8) & 7;
            d2 = c & 7;
            fprintf (out, "\\0%d%d%d", d0, d1, d2);
            col+=5;
          }
        }

        endp[0] = buf1; /* UNMODIFY resources */
	fprintf (out, "\"\n");
      }
    } else if ((r[0] == '\n') || (r[0] == '\r') || (r[0] == '#')) {
      /* skip comments and empty lines */
    } else {
      /* skip parts of other strings */
    }
    r = strchr(r,'\n');
    if (r == NULL) break;
    r++;
  }
  return 0;
}

int initResources (char * fileName) {
FILE * fp;
long ofs;
int c;
unsigned int i;

  resources = NULL;
  fp = fopen (fileName, "rb");
  fseek (fp, 0L, SEEK_END);
  ofs = ftell (fp);
  fseek (fp, 0L, SEEK_SET);
  resources = (char *) malloc (ofs);
  if (resources == NULL) return -__LINE__;
  for (i=0; i < ofs; i++) {
    c = fgetc (fp);
    if (c == EOF) {
      free (resources);
      resources = NULL;
      return -__LINE__;
    }
    resources[i] = c;
  }
  fclose (fp);
  return 0;
}

int main(int argc, char **argv) {
  FILE *out;
  unsigned int c;

  if (argc < 4) {
    fprintf(stderr, "Usage: strres <input bin file> <output C file> <prefix>\n");
    return 1;
  }

  c = initResources (argv[1]);
  if (c) {
    fprintf(stderr, "Cannot read input file (%s).\n", argv[1]);
    return 1;
  }

  if ((out = fopen (argv[2], "wt")) == NULL) {
    fprintf (stderr, "Cannot open output file (%s).\n", argv[2]);
    return 1;
  }
  fprintf (out, "#ifndef %sAUTOGEN_INCLUDED\n#define %sAUTOGEN_INCLUDED\n\n", argv[3], argv[3]);
  c = resourceTranslate (argv[3], out);
  fprintf (out, "\n#endif\n");
  fclose (out);

  return c;
}
